import enum


class ConditionalDescriptor:
    """
    A descriptor that allows class and instance dispatch to different methods.

    It receives a string name and dispatches to _<name> when accessed from
    an instance and _cls_<name> when accessed from a class.
    """

    def __init__(self, attr):
        self.attr = attr

    def __get__(self, instance, cls=None):
        if instance is None:
            return getattr(cls, '_cls_' + self.attr)
        else:
            return getattr(instance, '_' + self.attr)


class Namespace(enum._EnumDict):
    """
    Namespace that automatically create value() instances for member entries.
    """

    def __init__(self, data):
        super().__init__()
        self.update(data)
        self._last_values.extend(getattr(data, '_last_values', ()))
        self._member_names.extend(getattr(data, '_member_names', ()))
        self._auto_index = 0
        self._enum_values = {}

    def __setitem__(self, key, value):
        if key.isupper():
            if isinstance(value, str):
                value = TaggedInt(self._auto_index, value)
                self._auto_index += 1
            elif isinstance(value, tuple):
                value = TaggedInt(*value)
            elif isinstance(value, TaggedInt):
                self._auto_index = value + 1
            elif isinstance(value, int):
                value = TaggedInt(value, key.lower().replace('_', ' '))
                self._auto_index = value + 1
            # Proxy strings
            elif value == str(value):
                value = TaggedInt(self._auto_index, str(value))
                self._auto_index += 1
            else:
                type_name = value.__class__.__name__
                raise TypeError('unsupported enum value type: %s' % type_name)

            self._enum_values[key] = value

        super().__setitem__(key, value)

    def __getattr__(self, item):
        return getattr(self.data, item)


class DescriptionEnumMeta(enum.EnumMeta):
    """
    Metaclass for DescriptionMenu.
    """

    @classmethod
    def __prepare__(meta, cls, bases):  # noqa: N804
        namespace = super().__prepare__(cls, bases)
        return Namespace(namespace)

    def __init__(cls, name, bases, namespace):  # noqa: N805
        super().__init__(name, bases, namespace)

        # Handle enum values and store metadata
        enum_values = namespace._enum_values
        cls._descriptions = \
            {k: v.description for k, v in enum_values.items()}
        cls._descriptions.update(
            {i: i.description for i in enum_values.values()}
        )

        for k, v in enum_values.items():
            setattr(cls, k + '_DESCRIPTION', v.description)

    def _cls_get_description(cls, value):  # noqa: N805
        """
        Get description from value.
        """
        try:
            return cls._descriptions[value]
        except KeyError:
            raise ValueError('not a member of enumeration: %r' % value)


class IntEnum(enum.IntEnum, metaclass=DescriptionEnumMeta):
    """
    A subclass of enum.IntEnum that accepts an optional human-friendly
    description field during declaration.

    It is safe to translate description strings.

    Usage:
        >>> class Roles(IntEnum):
        ...     TEACHER = 0, 'teacher'
        ...     STUDENT = 1, 'student'

    """

    get_description = ConditionalDescriptor('get_description')

    def _get_description(self):
        """
        Return description string for value member.
        """
        return self.__class__._descriptions[int(self)]


class Enum(enum.Enum, metaclass=DescriptionEnumMeta):
    """
    Similar to :cls:`boogie.IntEnum`, but accepts any type of value.
    """

    get_description = ConditionalDescriptor('get_description')

    def _get_description(self):
        """
        Return description string for value member.
        """
        return self.__class__._descriptions[int(self)]


class TaggedInt(int):
    """
    A tagged integer.
    """

    def __new__(cls, value=0, description=''):
        self = super().__new__(cls, value)
        self.description = description
        return self

    def __repr__(self):
        name = self.__class__.__name__
        return '%s(%s, %r)' % (name, super().__repr__(), self.description)

    __str__ = __repr__



from ..utils.text import humanize_name


def join_url(head, *args):
    """
    Join url parts. It prevents duplicate backslashes when joining url
    elements.
    """
    if not args:
        return head
    else:
        tail = join_url(*args)
        return f"{head.rstrip('/')}/{tail.lstrip('/')}"


def validation_error(err, status_code=403):
    """
    Return a JSON message describing a validation error.
    """
    errors = err.messages
    msg = {'status_code': status_code, 'error': True}
    if len(errors) == 1:
        msg['message'] = errors[0]
    else:
        msg['messages'] = errors
    return msg


def natural_base_url(model):
    """
    Return the natural base url name for the given model:
    * Uses a plural form.
    * Convert CamelCase to dash-case.
    """
    name = model._meta.verbose_name_plural or model._meta.model_name
    return humanize_name(name).replace(' ', '-')



import re

RE_NAME = re.compile(r'^\w+$')
FULL_SLICE = slice(None, None, None)
INIT_TEMPLATE = """
def __init__(self, {fields}):
    self._data = [{fields}]
"""


def linear_namespace(name, fields):
    """
    Construct a mutable namedtuple-like type.

    Args:
        name (str):
            The new type name
        fields:
            A sequence of names for each value of the linear namespace.
    """

    fields = tuple(fields)
    prop = make_item_accessor
    class_dict = dict(
        __init__=make_init(fields),
        _fields=fields,
    )
    class_dict.update((name, prop(i)) for i, name in enumerate(fields))
    return type(name, (LinearNamespaceBase,), class_dict)


def linear_namespace_from_sequence(cls, data, copy=True):
    """
    Create a new linear space from a sequence

    If a list is given, and copy=False, it will be reused as data.

    >>> Point = linear_namespace('Point', ['x', 'y'])
    >>> linear_namespace_from_sequence(Point, [1, 2])
    """

    data = list(data)
    if len(data) != len(cls._fields):
        n = len(cls._fields)
        raise ValueError('expected a sequence with %s paramenters' % n)

    new = object.__new__(cls)
    new._data = list(data) if copy or type(data) is not list else data
    return new


def make_init(args):
    """
    Make init function for linear namespace from list of args.
    """

    invalid = [name for name in args if not is_valid_python_name(name)]
    if any(invalid):
        raise ValueError('invalid names: %s' % invalid)
    if len(set(args)) != len(args):
        raise ValueError('arguments cannot be repeated')

    names = ', '.join(args)
    ns = {}
    exec(INIT_TEMPLATE.format(fields=names), ns)
    return ns['__init__']


def make_item_accessor(idx):
    """
    Returns a property that mirrors access to the idx-th value of an object.
    """

    @property
    def attr(self):
        return self[idx]

    @attr.setter
    def attr(self, value):
        self[idx] = value

    return attr


class LinearNamespaceBase:
    """
    Linear namespace are used as the base class for Row() instances
    in Boogie query sets.
    """

    __slots__ = ('_data',)

    fromseq = classmethod(linear_namespace_from_sequence)

    def __setitem__(self, i, value):
        if i == FULL_SLICE and len(value) != len(self):
            raise ValueError('cannot')
        else:
            self._data[i] = value

    def __getitem__(self, i):
        return self._data.__getitem__(i)

    def __repr__(self):
        args = ', '.join(repr(x) for x in self)
        return '%s(%s)' % (type(self).__name__, args)

    # Sequence
    __iter__ = (lambda self: self._data.__iter__())
    __len__ = (lambda self: self._data.__len__())

    # Arithmetic
    __add__ = (lambda self, x: self._data.__add__(as_data(self, x)))
    __radd__ = (lambda self, x: as_data(self, x).__add__(self._data))
    __mul__ = (lambda self, x: self._data.__mul__(as_data(self, x)))
    __rmul__ = (lambda self, x: self._data.__rmul__(as_data(self, x)))

    # Logical operators
    __eq__ = (lambda self, x: self._data.__eq__(as_data(self, x)))
    __ne__ = (lambda self, x: self._data.__ne__(as_data(self, x)))
    __lt__ = (lambda self, x: self._data.__lt__(as_data(self, x)))
    __le__ = (lambda self, x: self._data.__le__(as_data(self, x)))
    __gt__ = (lambda self, x: self._data.__gt__(as_data(self, x)))
    __ge__ = (lambda self, x: self._data.__ge__(as_data(self, x)))

    # Public list methods that do not change the instance size
    count = (lambda self, value: self._data.count(value))
    index = (lambda self, value, *args: self._data.index(value, *args))
    sort = (lambda self, **kwargs: self._data.sort(**kwargs))
    reverse = (lambda self: self._data.reverse())


LINEAR_NAMESPACE_API = {
    '_fields', 'fromseq', 'index', 'count', 'sort', 'reverse',
}
LINEAR_NAMESPACE_CACHE = {}


def linear_namespace_cached(name, attrs):
    """
    A cached version of linear_namespace.

    This function will always return the same type for each unique combination
    of name, attrs.
    """

    attrs = tuple(attrs)
    try:
        return LINEAR_NAMESPACE_CACHE[name, attrs]
    except KeyError:
        new_type = linear_namespace(name, attrs)
        return LINEAR_NAMESPACE_CACHE.setdefault((name, attrs), new_type)


linear_namespace.fromseq = linear_namespace_from_sequence


#
# Utility functions
#
def is_valid_python_name(name):
    if not name or name[0].isdigit() or name.startswith('__'):
        return False
    if name in LINEAR_NAMESPACE_API:
        return False
    return RE_NAME.match(name) is not None


def as_data(self, other):
    """
    Convert other to a compatible data type.
    """
    if isinstance(other, list):
        return other
    elif isinstance(other, type(self)):
        return other._data
    elif isinstance(other, tuple):
        return list(other)
    return other



import base64
from hashlib import md5


def secret_hash(data):
    strings = []
    for key, value in sorted(data.items()):
        strings.append(key)
        try:
            data = hash(value)
            if data != -1:
                strings.append(str(data))
        except TypeError:
            pass
    print(strings)
    data = ''.join(strings)
    hash_value = md5(data.encode('utf8')).digest()
    return base64.b85encode(hash_value).decode('ascii')



def humanize_name(name):
    """
    "Humanize" camel case or Python variable name.

    Usage:
        >>> humanize_name('SomeName')
        'some name'
    """
    name = name.replace('_', ' ')
    parts = []
    buffer = []
    is_last_lower = False
    for chr in name:
        if chr.isupper():
            if is_last_lower:
                parts.append(''.join(buffer))
                buffer = [chr]
            else:
                buffer.append(chr)
            is_last_lower = False
        else:
            is_last_lower = True
            buffer.append(chr)

    parts.append(''.join(buffer))
    return '-'.join(parts)


def plural(st):
    """
    Convert string into a probable plural form.
    """
    return st + 's'



import logging

from django.core.exceptions import ImproperlyConfigured
from django.urls import path, include
from django.utils.functional import cached_property
from rest_framework import routers

from .api_info import ApiInfo
from .resource_info import ResourceInfo
from .router_builder import RouterBuilder

log = logging.getLogger('boogie.rest_api')


class RestAPI:
    router_class = routers.DefaultRouter

    @cached_property
    def last_version(self):
        versions = set(self.registries)
        versions.remove(None)
        return max(versions) if versions else 'v1'

    @cached_property
    def urls(self):
        versions = set(self.registries)
        versions.remove(None)
        if not versions:
            versions = ['v1']

        urlpatterns = []
        for version in sorted(versions):
            router = self.get_router(version)
            urlpatterns.append(
                path(version + '/', include(router.urls))
            )
        urlpatterns.append(
            path('', api_root_view(versions))
        )

        return urlpatterns

    def __init__(self):
        self.registries = {None: {}}

    def __call__(self, *args, **kwargs):
        if args and isinstance(args[0], type):
            return self(*args[1:], **kwargs)(args[0])

        def decorator(cls):
            self.register(cls, *args, **kwargs)
            return cls

        return decorator

    def register(self, model, fields=None, *, version=None, **kwargs):
        """
        Register class with the given meta data.

        Args:
            model:
                A Django model
            version:
                Optional API version string (e.g., 'v1'). If not given, it will
                register a resource to all API versions.
            fields:
                The list of fields used in the API. If not given, uses all
                fields.
            exclude:
                A list of fields that should be excluded.
            base_url:
                The base url address in which the resource is mounted. Defaults
                to a dashed case plural form of the model name.
            base_name:
                Base name for the router urls. Router will append suffixes such
                as <base_name>-detail or <base_name>-list. Defaults
                to a dashed case plural form of the model name.


        Returns:
            An EndPointInfo object.
        """
        info = ResourceInfo(model, fields, **kwargs)
        self.registries[version][model] = info

    def get_router(self, version):
        """
        Gets a DRF router object for the given API version.

        Args:
            version: An API versioning string.
        """
        info = api_info(self, version)
        router = self.router_class()
        router.root_view_name += '-' + version
        for resource in info.resources:
            builder = RouterBuilder(info, resource)
            builder.register_at(router)
        return router

    def get_urls(self, version):
        """
        Return a list of urls to be included in Django's urlpatterns::

        Usage:
            .. code-block:: python

                urlpatterns = [
                    ...,
                    path('api/v1/', rest_api.get_urls('v1'))
                ]

        See Also:
            :meth:`get_router`
        """
        return self.get_router(version).urls

    #
    # Decorators
    #
    def action(self, model, name=None, version=None):
        """
        Decorator that register a function as an action to the provided
        model.

        Args:
            model:
                A Django model or a string with <app_label>.<model_name>.
            name:
                The action name. It is normally derived from the action function
                by simply replacing underscores by dashes in the function
                name.

        Usage:

            .. code-block:: python

                @rest_api.action('auth.User')
                def books(user):
                    return user.user.books.all()
        """

    def property(self, model, func=None, *, version=None, name=None):
        """
        Decorator that declares a read-only API property.

        Args:
            model:
                The model name.
            version:
                API version. If ommited, it will be included in all API
                versions.
        """
        info = self.get_model_info(model, version)

        def decorator(func):
            nonlocal name

            name = name or func.__name__
            info.add_property(name, func)
            return func

        if func is not None:
            decorator(func)
            return

        return decorator

    def get_model_info(self, model, version=None):
        registry = self.registries.setdefault(version, {})
        try:
            return registry[model]
        except KeyError:
            model_name = model._meta.name
            version_string = '' if not version else ' (%s)' % version
            raise ImproperlyConfigured(
                '{model} is not registered on the API{version}. Please '
                'decorate the model class with the boogie.rest.rest_api() '
                'decorator.'.format(model=model_name, version=version_string)
            )


#
# Utility functions and classes
#

def api_info(rest_api: RestAPI, version: str) -> ApiInfo:
    """
    Builds an ApiInfo object.
    """
    registry_map = rest_api.registries.get(version, {})
    classes = {info.cls for info in registry_map.values()}
    for k, v in rest_api.registries[None].items():
        if v.model not in classes:
            registry_map[k] = v.copy()
    return ApiInfo(version, registry_map)


def api_root_view(versions, description=None):
    class RestApiView(routers.APIRootView):
        """
        This is the base entry point for REST resources and provides access to
        all versions of the REST API.
        """

    if description:
        RestApiView.__doc__ = description

    views = {version: 'api-root-' + version for version in versions}
    return RestApiView.as_view(api_root_dict=views)



import functools

from django.conf import settings
from django.core.exceptions import MiddlewareNotUsed, ImproperlyConfigured
from django.core.handlers.exception import get_exception_response, response_for_exception
from django.urls import get_resolver, get_urlconf
from django.utils.module_loading import import_string

from boogie.views.base import log
from .middleware import BOOGIE_VIEW_MIDDLEWARES


def not_implemented(request, **kwargs):
    raise NotImplementedError


def allowed_methods(view):
    """
    Return the default list of allowed HTTP methods for the view.
    """
    cls = type(view)
    all_methods = ['get', 'post', 'delete', 'put']
    implemented = (attr for attr in all_methods
                   if getattr(cls, attr) is not not_implemented)
    return ['options', *implemented]


def method_map(view):
    """
    Return a dictionary mapping HTTP method names to their corresponding
    view functions.
    """
    wrapper = view.wrap_method
    return dict(
        GET=wrapper(getattr(view, 'get', not_implemented)),
        POST=wrapper(getattr(view, 'post', not_implemented)),
        DELETE=wrapper(getattr(view, 'delete', not_implemented)),
        PUT=wrapper(getattr(view, 'put', not_implemented)),
        OPTIONS=wrapper(getattr(view, 'options', not_implemented)),
    )


def middleware_chain(middleware_list, handler):
    """
    Reduces the middleware chain into a single handler.
    """

    handler = safe_handler(handler)

    for ref in middleware_list:
        factory = load_middleware(ref)
        try:
            middleware = factory(handler)
        except MiddlewareNotUsed as exc:
            if settings.DEBUG:
                log.debug(f'MiddlewareNotUsed({exc}): {ref}')
            continue
        handler = safe_handler(middleware)

    return handler


def safe_handler(get_response):
    @functools.wraps(get_response)
    def inner(request):
        try:
            response = get_response(request)
        except TimeoutError as exc:
            log.warning(
                f'Timeout: {request.path}',
                extra={'status_code': 408, 'request': request},
            )
            resolver = get_resolver(get_urlconf())
            response = get_exception_response(request, resolver, 408, exc)
        except Exception as exc:
            response = response_for_exception(request, exc)
        return response

    return inner


def load_middleware(ref):
    """
    Loads a middleware from a string reference. Callable objects are just
    pass-thru.
    """

    if callable(ref):
        return ref
    elif not isinstance(ref, str):
        raise TypeError(f'invalid middleware reference: {ref.__class__}')
    if '.' in ref:
        return import_string(ref)
    try:
        return BOOGIE_VIEW_MIDDLEWARES[ref]
    except KeyError:
        raise ImproperlyConfigured(
            f'There is no middleware registered as {ref}. You can register, '
            f'new boogie view middlewares in the BOOGIE_VIEW_MIDDLEWARES '
            f'setting in the settings module.'
        )



class SocialAuthConf:
    # Social authentication
    SOCIAL_AUTH_LOGIN_REDIRECT_URL = '/'
    SOCIAL_AUTH_LOGIN_URL = '/auth/login/'

    # OAUTH keys (not working yet)
    SOCIAL_AUTH_GITHUB_KEY = ''
    SOCIAL_AUTH_GITHUB_SECRET = ''

    SOCIAL_AUTH_GOOGLE_OAUTH2_KEY = ''
    SOCIAL_AUTH_GOOGLE_OAUTH2_SECRET = ''
    SOCIAL_AUTH_GOOGLE_OAUTH2_SCOPE = ''

    SOCIAL_AUTH_FACEBOOK_KEY = '1085127354890672'
    SOCIAL_AUTH_FACEBOOK_SECRET = '9f6ba5c8721172acaf25a733c4c81a99'
    SOCIAL_AUTH_FACEBOOK_SCOPE = ['email']

    SOCIAL_AUTH_TWITTER_KEY = 'JrIVXbXZguPeUsnqcjtbQEWXH'
    SOCIAL_AUTH_TWITTER_SECRET = 'vMZjdO7DsUV8mVo46smQK2SHyhCxnXyc24gxH6J6cH08anWqHA'



from pathlib import Path

import environ

path_type = type(Path('path'))


def env(default, type=None, name=None, **kwargs):
    """
    Declare a value that can be overridden by environment variables.

    Usage:
        Used in class declarations:

        class Conf:
            THE_ANSWER = env(42)

        Now Conf instances have an THE_ANSWER attribute that has a default value
        of 42, but can be overridden by a "THE_ANSWER" environment variable.
        If the configuration class define a ``env_prefix`` attribute, it will be
        prepended to the name of the environment variable.

    Args:
        default:
            The default value for the variable, if not present in the
            environment.
        type:
            Variable type. Usually this can be inferred from the default value.
            Boogie also accept some special values described bellow.
        name:
            Explicit name of the environment variable. By default, the name is
            derived from ``<class>.env_prefix + <attribute_name>``.

    Notes:
        Boogie understands the following builtin types: str, bool, int, float,
        list, dict and tuple. It also understands the following pseudo-types:

        json:
            Environment holds arbitrary JSON data.
        path:
            Path to a valid file in the filesystem. Returns a
            :cls:`pathlib.Path` instance.
        db_url:
            URL to a database connection, e.g., sqlite:///db.sqlite3 or
            psql://user:password@domain:port/database
        cache_url:
            URL to cache provider, e.g., memcache://host:port
        search_url:
            URL to search provider, e.g., elasticsearch://host:port
        email_url:
            URL to e-mail provider, e.g., smtp+ssl://host:port

    Returns:
        A env variable descriptor
    """
    return EnvDescriptor(default, type=type, name=name, **kwargs)


def env_property(*args, default=None, type=None, name=None):
    """
    Like env(), but declares a property that receives the value of the
    environment variable (or None, if not set) and return the desired value.

    Usage:
        Used as a decorator:

        class Conf:
            @env.property
            def HEX_ANSWER(self, value):
                # The value of the environment variable is passed as the second
                # argument.
                if value is None:
                    return hex(42)
                else:
                    return hex(value)

    Args:
        The decorator can receive the following optional keyword arguments:
        default, type and name. They behave as the same arguments in
        :func:`boogie.configurations.env`.

    Notes:
        This function should be accessed as the "property" of the "env"
        function.

    See Also:
        :func:`boogie.configurations.env`

    Returns:
        A environ property.
    """
    kwargs = dict(default=default, type=type, name=name)

    # Return decorator
    if not args:
        return lambda func: env_property(func, **kwargs)

    fget, = args
    if default is None and type is None:
        kwargs.update(type=str)
    return EnvProperty(fget, **kwargs)


class EnvDescriptor:
    METHOD_MAPPER = {
        # Basic Python types
        str: 'str',
        bool: 'bool',
        float: 'float',
        int: 'int',
        list: 'list',
        tuple: 'tuple',
        dict: 'dict',
        path_type: 'str',

        # Special methods
        'json': 'json',
        'path': 'path',
        'db_url': 'db_url',
        'cache_url': 'cache_url',
        'search_url': 'search_url',
        'email_url': 'email_url',
    }

    def __init__(self, default, type=None, name=None, **kwargs):
        self.default = default
        self.name = name
        self.kwargs = kwargs

        if type is None:
            type = default.__class__
        if type not in self.METHOD_MAPPER:
            raise ValueError(f'invalid type: {type}')
        self.type = type

    def __set_name__(self, owner, name):
        if self.name is None:
            prefix = getattr(owner, 'env_prefix', '')
            self.name = prefix + name

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        value = self.get_value(instance)
        setattr(instance, self.name, value)
        return value

    def get_value(self, conf):
        method = self.get_env_method(conf)
        return method(self.name, default=self.default, **self.kwargs)

    def get_env_method(self, conf):
        method = self.METHOD_MAPPER[self.type]
        return getattr(conf.env, method)


class EnvProperty(EnvDescriptor):
    def __init__(self, fget, default=None, **kwargs):
        self.fget = fget
        super().__init__(default=default, **kwargs)

    def get_value(self, conf):
        value = super().get_value(conf)
        return self.fget(conf, value)


class Env(environ.Env):
    def __call__(self, name, default=None, type=None, **kwargs):
        if type is None and default is None:
            type = str
        elif type is None:
            type = default.__class__

        method_name = EnvDescriptor.METHOD_MAPPER[type]
        method = getattr(self, method_name)
        return method(name, default=default, **kwargs)


env.property = env_property



"""
Random Phrases
--------------

Generate human-friendly random strings. Phrases are usually of the form
"adjective" + "subjective". We try to use mostly characters from Science and
Science fiction movies.

The main entry point is the :func:`phrase` function, that simply returns a
random phrase. If you want to learn more, read the source :)

.. autofunction :: phrase
"""

import random

from faker import Factory

fake = Factory.create('en-us')

star_wars_characters = [
    'Han Solo', 'Darth Vader', 'C3PO', 'R2D2', 'Luke Skywalker',
    'Princess Leia', 'Jabba', 'Obi Wan', 'Yoda', 'Jar Jar Binks',
]
famous_scientists = [
    # Physicists
    'Einstein', 'Newton', 'Dirac', 'Bohr', 'Rutherford', 'Heisenberg',
    'Curie', 'Langevin', 'Boltzmann',

    # Mathematicians
    'Pythagoras', 'Peano', 'Hilbert', 'Gauss', 'Galois',

    # Computer science
    'Knuth', 'Turing', 'Tim',

    # Biology
    'Darwin', 'Mendel', 'Lamarck', 'Mayr', 'Dobzhansky',
]
adjective_list = [
    'grumpy', 'heroic', 'coward', 'brave', 'treacherous', 'powerful',
    'influential',
]
phrase_groups = []


def is_phase_provider(func):
    phrase_groups.append(func)
    return func


def phrase(maker=None):
    """
    A random easy to memorize phrase.
    """
    maker = random.choice(phrase_groups)
    return maker()


def phrase_lower(maker=None):
    """
    Like phrase, but normalize to lowercase results.
    """
    maker = random.choice(phrase_groups)
    return maker().lower()


def subjective_adjective_phrase(subjectives=None, adjectives=None):
    """
    Return a new subjective-adjective phrase such as "Grumpy Einstein" from a
    list of subjectives and adjectives.
    """
    subjective = random.choice(subjectives or famous_scientists)
    adjective = random.choice(adjectives or adjective_list)
    return '%s %s' % (adjective.title(), subjective)


@is_phase_provider
def random_star_wars_phrase():
    """
    Random phrase based on Star Wars ;-)
    """
    return subjective_adjective_phrase(star_wars_characters)


@is_phase_provider
def random_scientist_phrase():
    """
    Random phrase using important scientists.
    """
    return subjective_adjective_phrase(famous_scientists)


@is_phase_provider
def random_fake_phrase():
    """
    Use fake-factory names.
    """
    return subjective_adjective_phrase([fake.first_name()])


@is_phase_provider
def random_fake_catch():
    """
    Catch phrases
    """
    return fake.catch_phrase()



import logging
import os
import sys
from importlib import import_module

from django.core.exceptions import ImproperlyConfigured

from .descriptors import Env

log = logging.getLogger('boogie')


def save_configuration(conf_class, where=None):
    """
    Prepare Django to receive the current configurations. This function is
    usually called at the end of the configuration module and it creates an
    instance of the configuration class and injects all names in its namespace.

    Usage::

        from boogie.configurations import set_configuration, Default

        class Conf(Default):
            MY_OPTION = 42

        init_configuration(Conf)


    Args:
        conf_class:
            A configuration class.
        where:
            The module that it should inject configuration parameters. It can be
            a dictionary or a string with the module path.
    """
    if where is None:
        try:
            where = os.environ['DJANGO_SETTINGS_MODULE']
        except KeyError:
            raise ImproperlyConfigured(
                'You must either define the DJANGO_SETTINGS_MODULE environment '
                'variable or pass an explicit module name to the '
                'init_configuration function.'
            )
    if isinstance(where, str):
        where = import_module(where).__dict__
    elif isinstance(where, type(sys)):
        where = where.__dict__

    # We create an instance of the class configuration and import all symbols
    # with upper case names.
    try:
        conf = conf_class()
        where.update(conf.get_settings())
    except Exception as exc:
        log.error('Error loading configurations: %s' % exc)
        raise


class Conf:
    """
    Base class for all configuration classes.
    """

    ENVFILE = None
    env_prefix = ''

    @classmethod
    def save_settings(cls, where=None):
        """
        Calls :func:`save_configuration` with the current class.
        """
        save_configuration(cls, where)

    def __init__(self, **kwargs):
        self._settings = None
        self.env = Env()

        cls = type(self)
        for attr, value in kwargs.items():
            attr = attr.upper()
            if hasattr(cls, attr):
                setattr(self, attr, value)
            else:
                raise TypeError(f'invalid argument: {attr}')

    def prepare(self):
        """
        A hook that is executed once after initialization and before creating
        the default settings dictionary.

        The default implementation tries to read an env file if it is defined
        by the ENVFILE settings variable.
        """
        if self.ENVFILE:
            self.env.read_env(self.ENVFILE)

    def finalize(self, settings):
        """
        A hook that receives a settings dictionary and returns the final
        output of the get_settings() method.

        The default implementation simply pass the input dictionary forward.
        """
        return settings

    def get_settings(self):
        """
        Return a dictionary with all settings defined by the configuration.

        This method is idempotent. It caches the resulting settings and always
        return the same settings dictionary.
        """
        if self._settings is None:
            self.prepare()
            settings = {
                attr: getattr(self, attr, None)
                for attr in dir(self)
                if (attr.isupper() and not attr.startswith('_'))
            }
            if settings.get('ENVFILE') is None:
                settings.pop('ENVFILE', None)
            self._settings = self.finalize(settings)
        return dict(self._settings)



from django.db.models import Field, Lookup, Q
from django.db.models.sql import Query


#
# In order to support the __ne query, we monkey patch Django's Query class to
# accept the __ne suffix. This will make expressions such as "F.value != value"
# work both in boogie and in the vanilla queryset API.
#
@Field.register_lookup
class NotEqual(Lookup):
    lookup_name = 'ne'

    def as_sql(self, *args):
        lhs, lhs_params = self.process_lhs(*args)
        rhs, rhs_params = self.process_rhs(*args)
        return '%s <> %s' % (lhs, rhs), lhs_params + rhs_params


def patch_query(query_class=Query):
    """
    Patches Query class to accept 'ne' lookups
    """
    # Build lookup
    build_lookup_original = query_class.build_lookup

    def build_lookup(self, lookups, lhs, rhs):
        if rhs is None and lookups[-1:] == ['ne']:
            rhs, lookups[-1] = False, 'isnull'
        return build_lookup_original(self, lookups, lhs, rhs)

    query_class.build_lookup = build_lookup

    # Resolve lookup
    if hasattr(query_class, 'resolve_lookup_value'):
        return patch_query_with_resolve_lookup_value(query_class)


def patch_query_with_resolve_lookup_value(query_class=Query):
    resolve_lookup_value_original = query_class.resolve_lookup_value

    def resolve_lookup_value(self, value, lookups, *args):
        if value is None and lookups[-1:] == ['ne']:
            value, lookups[-1] = False, 'isnull'
        return resolve_lookup_value_original(self, value, lookups, *args)

    query_class.prepare_lookup_value = resolve_lookup_value


#
# Utility methods
#
def lookup_method(lookup):
    "Factory function for lookup methods."

    def method(self, value):
        key = '%s__%s' % (self._name, lookup)
        return Q(**{key: value})

    method.__doc__ = 'Performs an name__%s lookup' % lookup
    method.__name__ = lookup
    return method


def lookup_property(lookup):
    "Factory function for simple lookup properties."

    def fget(self):
        return type(self)('%s__%s' % (self._name, lookup))

    return property(fget)


# Execute monkey patch
patch_query()



from functools import wraps

from .manager import QueryManager


def manager_method(method):
    """
    Decorator that marks a method as a manager method, i.e., it can only be
    accessed by a manager created with QuerySet.as_manager() and not by
    instances of the queryset itself.
    """

    @wraps(method)
    def decorated(self, *args, **kwargs):
        raise TypeError('this method cannot be used from a queryset.')

    decorated._is_manger_method = True
    decorated._original_method = method

    return decorated


def manager_property(method=None, *, classproperty=False):
    """
    Decorates a method that returns a QuerySet instance and expose it as a
    manager. This is useful to define custom accessors for related value,
    providing a consistent api between your custom managers and the related
    managers automatically created by Django.

    Usage:

        .. code-block:: python

            # models.py

            class Book(models.Model):
                authors = models.ManyToManyField('User')
                ...

                @manager_property
                def active_authors(self):
                    return self.authors.filter(active=True)


        The ``active_authors`` property is a manager, not a queryset. While most
        manager methods are available to query sets, making them interchangeable
        in many situations, exposing the property as actual managers
        improves the API consistency so users don't have to memorize which
        methods are present in requires calls to .all() and which ones do not.

        >>> [author.name for author in book.active_authors.all()]
        ['Douglas Adams', ...]
    """

    if method is None:
        return lambda f: manager_property(f, classproperty=classproperty)

    class ManagerProperty(property):
        def __get__(self, instance, cls=None):
            if instance is not None:
                return super().__get__(instance, cls)
            elif not classproperty:
                return self
            else:
                return self.fget(cls)

    return ManagerProperty(lambda self: QueryManager(method(self)))



import collections
from functools import lru_cache, partial

from django.db import models
from django.db.models.query import ValuesListIterable
from manager_utils import ManagerUtilsQuerySet
from django_pandas.managers import DataFrameQuerySet
from sidekick import lazy

from .expressions import F
from ..types import linear_namespace

F_EXPR_TYPE = type(F.age > 0)


class QuerySet(ManagerUtilsQuerySet, DataFrameQuerySet):
    """
    Boogie's drop in replacement to Django's query sets.

    It extends the query set API with a Pydata-inspired interface to select data
    """

    # Properties
    index = property(lambda self: Index(self))
    __column_names = None
    __is_column_slice = False

    # Class methods
    def as_manager(cls):
        # Overrides to use Boogie manager instead of the default manager

        from .manager import Manager
        manager = Manager.from_queryset(cls)()
        manager._built_with_as_manager = True
        return manager

    as_manager.queryset_only = True
    as_manager = classmethod(as_manager)

    def __getitem__(self, item):
        # 1D indexing is basic delegated to Django. We extend it with some
        # idioms that preserve backwards compatilibity
        if not isinstance(item, tuple):
            return getitem_1d(self, item)

        try:
            row, col = item
        except IndexError:
            raise TypeError('only 1d or 2d indexing is allowed')

        return getitem_2d(self, row, col)

    def __setitem__(self, item, value):
        # Similarly to getitem, we dispatch for the 1d and 2d indexing
        # functions
        if not isinstance(item, tuple):
            return setitem_1d(self, item, value)

        try:
            row, col = item
        except IndexError:
            raise TypeError('only 1d or 2d indexing is allowed')

        return setitem_2d(self, row, col, value)

    # Better accessors
    def update_item(self, pk, **kwargs):
        """
        Updates a single item in the queryset.
        """
        return self.filter(pk=pk).update(**kwargs)

    update_item.alters_data = True

    #
    # Enhanced API
    #
    def select_columns(self, *fields):
        """
        Similar to .values_list(*fields)
        """
        return select_columns(self, list(fields))

    def values(self, *fields, **expressions):
        qs = super().values(*fields, **expressions)
        fields = fields + tuple(expressions.keys())
        return self._mark_column_names(fields, qs)

    def values_list(self, *fields, **kwargs):
        qs = super().values_list(*fields, **kwargs)
        return self._mark_column_names(fields, qs)

    def _mark_column_names(self, columns, qs=None):
        qs = self if qs is None else qs
        qs.__is_column_slice = True
        qs.__column_names = columns
        return qs

    #
    # Pandas data frame and numpy array APIs
    #
    def to_dataframe(self):
        """
        Convert query set to a Pandas data frame
        """
        import pandas as pd

        if self.__column_names:
            fields = self.__column_names
        else:
            fields = self.model._meta.fields

        # Build data frame
        data = list(self.values_list('pk', *fields))
        df = pd.DataFrame(data, columns=['Index[pk]', *fields])
        df.index = df.pop('Index[pk]')
        df.index.name = 'pk'
        return df

    def update_dataframe(self, dataframe, batch_size=None, in_bulk=True):
        """
        Persist data frame data to the database. Data frame index must
        correspond to primary keys.

        Args:
            dataframe:
                A pandas data frame
            in_bulk (bool):
                If True (default), save values in bulk.
            batch_size:
                If saving in bulk, defines the size of each batch that touches
                the database. This avoids creating very long SQL commands that
                can halt the database for a perceptible amount of time.
        """
        objects = []
        fields = dataframe.columns
        add_object = objects.append
        new_object = self.model

        for pk, row in zip(dataframe.index, dataframe.to_dict('records')):
            row.setdefault('pk', pk)
            add_object(new_object(**row))

        if in_bulk:
            self.bulk_update(objects,
                             update_fields=fields,
                             batch_size=batch_size)
        else:
            for obj in objects:
                obj.save(update_fields=fields)

    def head(self, n=5):
        """
        Select the first n rows in the query set.
        """
        return self[:n]

    def tail(self, n=5):
        """
        Select the last n rows in the query set.
        """
        return self[-n:]


#
# Auxiliary classes
# ----------------------------------------------------------------------------
class QuerysetSequence(collections.Sequence):
    """
    Base class for sequences tyeps fed by query sets instances.
    """

    @staticmethod
    def prepare(queryset):
        return queryset

    _prepared = lazy(lambda self: self.prepare(self.queryset))

    def __init__(self, queryset):
        self.queryset = queryset

    def __iter__(self):
        return iter(self._prepared)

    def __len__(self):
        return len(self._prepared)

    def __getitem__(self, item):
        return self._prepared[item]

    def __repr__(self):
        return '%s(%s)' % (type(self).__name__, list(self))


class Index(QuerysetSequence):
    """
    Returns an index with the primary keys of all elements.
    """

    @staticmethod
    def prepare(qs):
        return qs.values_list('pk', flat=True)


class RowIterable(ValuesListIterable):
    """
    Iterable returned by QuerySet.values_list(named=True) that yields a
    namedtuple for each row.
    """

    @staticmethod
    @lru_cache()
    def _create_row_class(names):
        return linear_namespace('Row', names)

    def __init__(self, *args, **kwargs):
        self.field_names = kwargs.pop('field_names', None)
        super().__init__(*args, **kwargs)

    def __iter__(self):
        queryset = self.queryset
        if self.field_names:
            names = self.field_names
        elif queryset._fields:
            names = queryset._fields
        else:
            query = queryset.query
            names = list(query.extra_select)
            names.extend(query.values_select)
            names.extend(query.annotation_select)

        new = self._create_row_class(tuple(names))
        for row in super().__iter__():
            yield new(*row)


#
# Getitem auxiliary functions
# ----------------------------------------------------------------------------
def getitem_1d(qs, item):
    """
    Must dispatch to super() in all situations that Django supports:

    We keep Django's behavior in all of the following cases (a, b >= 0):
    * qs[a]
    * qs[a:b]

    Small extensions:
    * qs[-a] -> fetch from last
    * qs[a, -b] -> specify a index from first to last

    Larger extensions:
    * qs[list] -> ...
    """

    # Integer indexes
    if isinstance(item, int):
        if item >= 0:
            return super(QuerySet, qs).__getitem__(item)
        else:
            idx = -item - 1
            return qs.reverse()[idx] if qs.ordered else qs.order_by('-id')[idx]

    # Slicing
    if isinstance(item, slice):
        start, stop, step = item.start, item.stop, item.step
        if step is None and \
                (start is None or start >= 0) and \
                (stop is None or stop >= 0):
            return super(QuerySet, qs).__getitem__(item)

    raise TypeError('invalid index type: %r' % item.__class__.__name__)


def getitem_2d(qs, rows, cols):
    """
    Extract elements from queryset possibly filtering specific rows
    or columns.

    The first index selects rows and the second columns.
    """

    # First we filter the rows
    if is_vector_row_access(rows):
        qs = filter_queryset(qs, rows)
        return select_columns(qs, cols)
    else:
        qs = qs.filter(pk=rows)
        return select_columns(qs, cols)[0]


def is_vector_row_access(row):
    """
    Return True to all types that select multiple rows.
    """
    return isinstance(row, (list, slice, QuerySet, F_EXPR_TYPE))


def filter_queryset(qs, rows):
    """
    Extended ways to filter a queryset.
    """

    # Filter by pk
    if isinstance(rows, list):
        model = qs.model
        pks = {x.pk if isinstance(x, model) else x for x in rows}
        return qs.filter(pk__in=pks)

    elif isinstance(rows, slice):
        if rows == slice(None, None, None):
            return qs
        raise IndexError('invalid column slice: %s' % rows)

    elif isinstance(rows, QuerySet):
        return qs & rows

    elif isinstance(rows, F_EXPR_TYPE):
        return qs.filter(rows)

    raise TypeError('invalid row selector: %r' % rows.__class__.__name__)


def select_columns(qs, cols):
    if isinstance(cols, list):
        qs = qs.values_list(*cols, flat=False)
        field_names = get_column_names(cols, qs)
        qs._iterable_class = partial(RowIterable, field_names=field_names)
        return qs

    elif isinstance(cols, slice):
        if cols == slice(None, None, None):
            return qs
        raise IndexError('invalid column slice: %s' % cols)

    elif isinstance(cols, (str, F, models.F)):
        return qs.values_list(cols, flat=True)

    raise TypeError('invalid column selector: %r' % cols.__class__.__name__)


def get_column_names(cols, qs):
    """
    Return a list of valid field names for the given columns.
    """
    names = []
    append = names.append
    for col in cols:
        if isinstance(col, str):
            name = col
        elif isinstance(col, F):
            name = col._name
        else:
            raise TypeError('invalid column', col)
        append(name)
    return names


#
# Setitem auxiliary functions
# ----------------------------------------------------------------------------
def setitem_1d(qs, item, value):
    raise TypeError('invalid index type: %r' % item.__class__.__name__)


def setitem_2d(qs: QuerySet, rows, cols, value):
    if is_vector_row_access(rows):
        raise NotImplementedError
    elif isinstance(cols, str):
        return qs.update_item(rows, **{cols: value})
    elif isinstance(cols, list):
        cols = map(as_col_name, cols)
        mapping = dict(zip(cols, value))
        return qs.update_item(rows, **mapping)
    raise NotImplementedError


def set_columns(qs: QuerySet, cols, value):
    raise TypeError('invalid column selector: %r' % cols.__class__.__name__)


def as_col_name(obj):
    "Coerce object to a column name"

    if isinstance(obj, str):
        return obj
    elif isinstance(obj, F):
        print(obj, obj.__dict__)
        raise ValueError('invalid column object: %s' % obj)
    return obj.name



import datetime

from django.apps import apps
from django.contrib.auth.models import (
    AbstractBaseUser,
    PermissionsMixin,
    BaseUserManager,
)
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db import transaction
from django.urls import reverse
from django.utils import timezone
from django.utils.text import slugify
from django.utils.translation import ugettext_lazy as _, ugettext as __
from sidekick import delegate_to, lazy

from ...functions import get_config

strptime = datetime.datetime.strptime


class UserManager(BaseUserManager):
    """
    A manager that mimics the interface of Django's default User manager.
    """

    def create_user(self, *args, username=None, commit=True, **kwargs):
        if username is not None:
            kwargs['alias'] = username
        new = self.model(**kwargs)
        new.set_password(kwargs.get('password'))
        if commit:
            new.save()
        return new

    def create_superuser(self, *args, commit=True, **kwargs):
        user = self.create_user(*args, commit=False, **kwargs)
        user.is_staff = user.is_superuser = True
        if commit:
            user.save()
        return user

    def get_by_natural_key(self, email):
        return self.get(email=email)


class User(AbstractBaseUser, PermissionsMixin):
    """
    Base user model.
    """

    USERNAME_FIELD = 'email'
    email = models.EmailField(
        _('E-mail'),
        db_index=True,
        unique=True,
        help_text=_(
            'Users can register additional e-mail addresses. This is the '
            'main e-mail address which is used for login.'
        )
    )
    name = models.CharField(
        _('Name'),
        max_length=140,
        help_text=_(
            'Full name of the user.'
        )
    )
    alias = models.CharField(
        _('Alias'),
        max_length=20,
        help_text=_(
            'Public alias used to identify the user.'
        )
    )
    is_staff = models.BooleanField(
        _('staff status'),
        default=False,
        help_text=_(
            'Designates whether the user can log into the admin site.'
        ),
    )
    is_active = models.BooleanField(
        _('active'),
        default=True,
        help_text=_(
            'Designates whether this user should be treated as active. '
            'Unselect this instead of deleting accounts.'
        ),
    )
    date_joined = models.DateTimeField(
        _('date joined'),
        default=timezone.now
    )

    objects = UserManager()

    # Properties defined for better compatibility with default user model
    username = property(lambda self: self.alias)
    first_name = property(lambda self: self.partition(' ')[0])
    last_name = property(lambda self: self.partition(' ')[-1])
    profile_class = None

    class Meta:
        abstract = True

    def __str__(self):
        return self.name

    @lazy
    def profile(self):
        profile_class = self.profile_class
        if profile_class is None:
            try:
                profile_class = apps.get_model(self._meta.app_label, 'Profile')
            except KeyError:
                return None

        if self.id is None:
            return profile_class(user=self)
        else:
            try:
                return profile_class.objects.get(user=self)
            except ObjectDoesNotExist:
                return profile_class.objects.create(user=self)

    def save(self, *args, **kwargs):
        new = self.id is None
        if self.alias is None:
            self.alias = slugify(self.name)

        if new:
            with transaction.atomic():
                super().save(*args, **kwargs)
                if self.profile is not None:
                    self.profile.save()
        else:
            super().save(*args, **kwargs)

    def natural_key(self):
        return 'email'

    def get_full_name(self):
        return self.name.strip()

    def get_short_name(self):
        return self.alias

    def get_absolute_url(self):
        return reverse('users:profile-detail', args=(self.id,))


class Profile(models.Model):
    """
    Social information about users.
    """

    user = models.OneToOneField(
        get_config('AUTH_USER_MODEL', 'auth.User'),
        on_delete=models.CASCADE,
        verbose_name=_('user'),
        related_name='profile_ref',
    )

    # Delegates and properties
    username = delegate_to('user', readonly=True)
    name = delegate_to('user', readonly=True)
    first_name = delegate_to('user', readonly=True)
    last_name = delegate_to('user', readonly=True)
    alias = delegate_to('user', readonly=True)
    email = delegate_to('user', readonly=True)

    class Meta:
        abstract = True

    def __str__(self):
        if self.user is None:
            return __('Unbound profile')
        full_name = self.user.get_full_name() or self.user.username
        return __('%(name)s\'s profile') % {'name': full_name}

    def get_absolute_url(self):
        if self.user is not None:
            return self.user.get_absolute_url()



from ..base import Conf


class UserenaConf(Conf):
    USERENA_ACTIVATION_REQUIRED = False
    USERENA_SIGNIN_AFTER_SIGNUP = True
    USERENA_DISABLE_PROFILE_LIST = True
    USERENA_ACTIVATION_DAYS = 7
    USERENA_FORBIDDEN_USERNAMES = (
        'signup', 'signout', 'signin', 'activate', 'me', 'password', 'login',
        'codeschool'
    )
    USERENA_USE_HTTPS = False
    USERENA_REGISTER_PROFILE = False
    USERENA_SIGNIN_REDIRECT_URL = '/'
    USERENA_REDIRECT_ON_SIGNOUT = '/'
    USERENA_PROFILE_LIST_TEMPLATE = 'auth/profile-list.jinja2'



from collections import Mapping

from django.utils.functional import cached_property


class ApiInfo(Mapping):
    """
    Stores information about all resources of an specific API version.
    """

    @cached_property
    def resources(self):
        return list(self.registry.values())

    def __init__(self, version: str, registry: dict = None):
        self.version = version
        self.registry = dict(registry or {})

    def __getitem__(self, item):
        return self.registry[item]

    def __len__(self):
        return len(self.registry)

    def __iter__(self):
        return iter(self.registry)

    def get_base_name(self, model):
        """
        Return the base_name string for the given model.
        """
        resource = self.registry[model]
        return '%s-%s' % (self.version, resource.base_name)

    def get_lookup_field(self, model):
        """
        Return the default lookup_field used for the given model.
        """
        return 'pk'



from sidekick import lazy


class GeoAttribute:
    """
    F.geo.* namespace.
    """

    @lazy
    def _gis(self):
        import django.contrib.gis.db.models.functions as functions
        return functions

    def __init__(self, f_expr):
        self.expr = f_expr
        self.name = f_expr.name

    def __repr__(self):
        return 'F.%s.geo' % self.name

    # Gis properties
    area = property(lambda self: self._gis.Area)
    centroid = property(lambda self: self._gis.Centroid(self))
    envelope = property(lambda self: self._gis.Envelope(self))
    mem_size = property(lambda self: self._gis.MemSize(self))
    num_geometries = property(lambda self: self._gis.NumGeometries(self))
    num_points = property(lambda self: self._gis.NumPoints(self))
    perimeter = property(lambda self: self._gis.Perimeter(self))
    point_on_surface = property(lambda self: self._gis.PointOnSurface(self))

    # TODO: Spatial lookups
    # contained = method('contained')
    # coveredby = method('coveredby')
    # covers = method('covers')
    # crosses = method('crosses')
    # disjoint = method('disjoint')
    # equals = method('equals')
    # intersects = method('intersects')
    # relate = starmethod('relate')
    # touches = method('touches')
    # left = method('left')
    # right = method('right')
    # above = method('strictly_above')
    # below = method('strictly_below')

    # TODO: Gis functions
    # geojson = method(self._gis.AsGeoJSON)
    # gml = method(self._gis.AsGML)
    # kml = method(self._gis.AsKML)
    # svg = method(self._gis.AsSVG)
    # bounding_circle = method(self._gis.BoundingCircle)
    # difference = method(self._gis.Difference)
    # distance = method(self._gis.Distance)
    # force_rhr = method(self._gis.ForceRHR)
    # geohash = method(self._gis.GeoHash)  # __hash__ requires an int
    # intersection = method(self._gis.Intersection)
    # make_valid = method(self._gis.MakeValid)
    # reverse = method(self._gis.Reverse)
    # scale = method(self._gis.Scale)
    # snap_to_grid = method(self._gis.SnapToGrid)
    # symmetric_difference = method(self._gis.SymDifference)
    # transform = method(self._gis.Transform)
    # translate = method(self._gis.Translate)
    # union = method(self._gis.Union)



import os
from pathlib import Path

from sidekick import unique

from .base import Conf
from .descriptors import env, env_property
from .tools import secret_hash


class BaseConf(Conf):
    """
    Base class for Django configurations. Users should use DjangoConf
    """

    env_prefix = 'DJANGO_'

    def finalize(self, settings):
        settings = super().finalize(settings)
        if settings['SECRET_KEY'] is None:
            settings['SECRET_KEY'] = secret_hash(settings)
        return settings

    #: Django base environment. We suggest distinguishing between 'local',
    #: 'test' and 'production'.
    ENVIRONMENT = env('local')

    #: WARNING: keep the secret key used in production secret! We generate a
    #: secret from a hash of the current settings during the .finalize() phase.
    #: this is ok for local development, but may be insecure/inconvenient for
    #: production.
    @env_property
    def SECRET_KEY(self, value):  # noqa: N802
        if value is None:
            if self.ENVIRONMENT in ('local', 'test'):
                return self.ENVIRONMENT
            else:
                return None
        return value

    #: Build paths inside the project like this: BASE_DIR / 'foo'
    @env_property
    def BASE_DIR(self, value):  # noqa: N802
        if value is None:
            conf_dir = Path(os.path.dirname(type(self).__file__))

            return conf_dir
        return value

    #: Location of the django project
    @property
    def django_project_path(self):
        name, _, _ = os.environ['DJANGO_SETTINGS_MODULE'].rpartition('.')
        return name

    #: By default, debug is enabled only on 'local' and 'test' environments.
    @env_property(type=bool)
    def DEBUG(self, value):  # noqa: N802
        if value is None:
            return self.ENVIRONMENT in ('test', 'local')
        return value

    ALLOWED_HOSTS = env([])

    # Database
    # https://docs.djangoproject.com/en/2.0/ref/settings/#databases
    DATABASE_DEFAULT = env('sqlite:///local/db/db.sqlite3', type='db_url', name='DJANGO_DEFAULT_DB')
    DATABASE_TEST = env('sqlite:///local/db/test.sqlite3', type='db_url', name='DJANGO_TEST_DB')
    DATABASES = property(lambda self: {
        'default': self.DATABASE_DEFAULT,
        'test': self.DATABASE_TEST
    })

    # Password validation
    # https://docs.djangoproject.com/en/2.0/ref/settings/#auth-password-validators
    AUTH_PASSWORD_VALIDATORS = [
        {
            'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
        },
        {
            'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
        },
        {
            'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
        },
        {
            'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
        },
    ]

    @env_property
    def ROOT_URLCONF(self, value):  # noqa: N802
        if value is None:
            return self.django_project_path + '.urls'
        return value

    @env_property
    def WSGI_APPLICATION(self, value):  # noqa: N802
        if value is None:
            return self.django_project_path + '.wsgi.application'
        return value

    # Internationalization
    # https://docs.djangoproject.com/en/2.0/topics/i18n/
    LANGUAGE_CODE = env('en-us')
    TIME_ZONE = env('UTC')
    USE_I18N = env(True)
    USE_L10N = env(True)
    USE_TZ = env(True)

    #: Static files (CSS, JavaScript, Images)
    #: https://docs.djangoproject.com/en/2.0/howto/static-files/
    STATIC_URL = env('/static/')

    #: It is often convenient to enable/disable Django ability to serve static
    #: files.
    SERVE_STATIC_FILES = env(True)

    #: Site admin url, if Django admin app is installed.
    ADMIN_URL = env('admin/')
    SITE_ID = 1

    #: Authentication
    LOGIN_URL = env('login/')
    LOGOUT_URL = env('logout/')
    USER_MODEL = 'auth.User'


class InstalledAppsConf(BaseConf):
    """
    Mixin class that helps managing INSTALLED_APPS.
    """

    INSTALLED_APPS = property(lambda self: self.get_installed_apps())  # noqa: N802

    disable_django_users = False
    disable_django_admin = False
    extra_contrib_apps = []
    third_party_apps = []
    project_apps = []

    def get_installed_apps(self):
        """
        Return a list of installed apps.

        The list contains project  apps + thirdy parties + contrib, in that
        order.
        """
        return list(unique([
            *self.get_project_apps(),
            *self.get_third_party_apps(),
            *self.get_contrib_apps(),
        ]))

    def get_project_apps(self):
        """
        Return a list of apps created specifically for the project.
        """
        return list(self.project_apps)

    def get_third_party_apps(self):
        """
        Return a list of third party dependencies.
        """
        return list(self.third_party_apps)

    def get_contrib_apps(self):
        """
        Return a list of apps from django.contrib.
        """

        apps = list(self.extra_contrib_apps)
        if not self.disable_django_admin:
            apps.append('django.contrib.admin')
        if not self.disable_django_users:
            apps.extend(['django.contrib.auth', 'django.contrib.sessions'])
        apps.extend(['django.contrib.contenttypes', 'django.contrib.messages'])
        if self.SERVE_STATIC_FILES:
            apps.append('django.contrib.staticfiles')
        return apps

    def with_app(self, app, app_list, deps=None, optdeps=None):
        """
        Insert app on list, respecting its dependencies.

        Usage:
            Useful for third parties declare dependency classes:

            class MyAppConf(InstalledAppsConf):
                def get_third_party_apps(self):
                    apps = super().get_third_party_apps()
                    return self.with_app('my_app', apps, deps=['dep1', 'dep2'])

        Args:
            app:
                App name.
            app_list:
                List of apps.
            deps:
                List of hard dependencies. All apps in that list are also
                inserted in the result just after the inserted app.
            optdeps:
                List of optional dependencies. App will be inserted just
                before all apps in list if they appear in the app list.

        Returns:
            A list of apps.
        """
        apps = [app]
        apps.extend(x for x in (deps or ()) if x not in app_list)
        all_deps = set(deps or ())
        all_deps.update(optdeps or ())

        result = []
        for app in app_list:
            if apps and app in all_deps:
                result.extend(apps)
                apps = None
            result.append(app)
        return result


class MiddlewareConf(InstalledAppsConf):
    """
    Configure middleware classes.
    """

    MIDDLEWARE = property(lambda self: self.get_middleware())
    with_middleware = InstalledAppsConf.with_app

    def get_middleware(self):
        apps = self.INSTALLED_APPS

        def add(middleware):
            if middleware.partition('.middleware')[0] in apps:
                return [middleware]
            else:
                return []

        return [
            'django.middleware.security.SecurityMiddleware',
            *add('django.contrib.sessions.middleware.SessionMiddleware'),
            'django.middleware.common.CommonMiddleware',
            'django.middleware.csrf.CsrfViewMiddleware',
            *add('django.contrib.auth.middleware.AuthenticationMiddleware'),
            *add('django.contrib.messages.middleware.MessageMiddleware'),
            'django.middleware.clickjacking.XFrameOptionsMiddleware',
        ]


class TemplatesConf(BaseConf):
    """
    Configure templates.
    """

    DJANGO_TEMPLATES_DIRS = ()
    JINJA_TEMPLATES_DIRS = ()

    def get_django_templates_dirs(self):
        return list(self.DJANGO_TEMPLATES_DIRS)

    def get_jinja_templates_dirs(self):
        return list(self.JINJA_TEMPLATES_DIRS)

    @property
    def TEMPLATES(self):  # noqa: N802
        templates = [self.DJANGO_TEMPLATES, self.JINJA_TEMPLATES]
        return [x for x in templates if x]

    @property
    def DJANGO_TEMPLATES(self):  # noqa: N802
        return {
            'BACKEND': 'django.template.backends.django.DjangoTemplates',
            'DIRS': self.get_django_templates_dirs(),
            'APP_DIRS': True,
            'OPTIONS': {
                'context_processors': [
                    'django.template.context_processors.debug',
                    'django.template.context_processors.request',
                    'django.contrib.auth.context_processors.auth',
                    'django.contrib.messages.context_processors.messages',
                ],
            },
        }

    @property
    def JINJA_TEMPLATES(self):  # noqa: N802
        return {
            'BACKEND': 'django.template.backends.jinja2.Jinja2',
            'DIRS': self.get_jinja_templates_dirs(),
            'APP_DIRS': True,
            'OPTIONS': {},
        }


class DjangoConf(MiddlewareConf,
                 TemplatesConf,
                 InstalledAppsConf,
                 BaseConf):
    """
    Base configuration class for Django-based projects.
    """



from django.db import models
from django.db.models import expressions
from django.db.models import functions

from .fix import lookup_method, lookup_property
from .functions import coalesce
from .geo import GeoAttribute


class FMeta(type(models.F)):
    def __getattr__(cls, attr):  # noqa: N805
        return cls(attr)


class F(expressions.Combinable, metaclass=FMeta):
    """
    Replacement for Django's F, Q, and many Func objects.

    Several use cases:

    * Direct value access:
        ``F.user <==> F('user')``
    * Sub-value access:
        ``F.user.name <==> F('user__name')``
    * Queries:
        ``F.user.created > date <==> Q(user__created__gte=date)``
    * SQL Functions:
        ``F.user.created.min() <==> Min('user__created')``
    """

    geo = property(lambda self: GeoAttribute(self))

    def __init__(self, name):
        self._name = name

    def __repr__(self):
        return 'F.%s' % (self._name)

    def __str__(self):
        return self._name

    def __getattr__(self, attr):
        return self.getattr(attr)

    # F-expression
    def _delegate_to_f_object(self, *args, **kwargs):
        fobj = models.F(self._name)
        return fobj.resolve_expression(*args, **kwargs)

    asc = desc = resolve_expression = _delegate_to_f_object
    default_alias = 'boogie.F'

    #
    #  Lookup operators not implemented by Combinable objects
    #
    def __eq__(self, other):
        return models.Q(**{self._name: other})

    __ne__ = lookup_method('ne')
    __lt__ = lookup_method('lt')
    __le__ = lookup_method('lte')
    __gt__ = lookup_method('gt')
    __ge__ = lookup_method('gte')
    __pos__ = asc
    __neg__ = desc
    __hash__ = lambda self: hash(self._name)

    def __len__(self):
        msg = 'len() function not supported, please use F.%s.length() instead.'
        raise TypeError(msg % self._name)

    #
    #  Generic functionality
    #
    def getattr(self, attr):
        """
        F.value.getattr('sub_field') <==> F.value.sub_field.

        This method is necessary for deeply nested sub-queries if attribute
        name conflicts with some method of the F class.
        """
        print(self, attr)
        return type(self)('%s__%s' % (self._name, attr))

    def cast(self, type):
        """
        Coerce an expression to a new value type.
        """
        return functions.Cast(self._name, type)

    def with_default(self, *args):
        """
        Return value, if not null, or the first non null argument.

        Arguments can be values or other value attributes.

        Usage:
            F.age.with_default(18)
            F.modified.with_default(F.created, now())
        """
        return coalesce(self, *args)

    def in_sequence(self, values):
        """
        Check if value is present in a sequence of values.

        If the passed values is a queryset, it performs a nested query. Nested
        queries can be either slower (typically) or faster than an explicit
        lists depending on the situation and backend.
        """
        return models.Q(**{self._name + '__in': values})

    def in_range(self, start, end):
        """
        Check if value is within the given range.
        """
        return models.Q(**{self._name + '__range': (start, end)})

    #
    # Statistics
    #
    def count(self, distinct=False):
        """
        Number of non-null occurrences of a value in the database.

        If distinct=True, only count distinct values.
        """
        return models.Count(self._name, distinct=distinct)

    def mean(self):
        """
        Mean value of value.
        """
        return setting_attrs(models.Avg(self._name), name='mean')

    def std(self, sample=False):
        """
        Standard deviation of value.

        If sample=True, return the sample standard deviation.
        """
        return models.StdDev(self._name, sample=sample)

    def var(self, sample=False):
        """
        Variance of value.

        If sample=True, return the sample variance.
        """
        return models.Variance(self._name, sample=sample)

    def min(self):
        """
        Minimum value of value.
        """
        return models.Min(self._name)

    def max(self):
        """
        Maximum value of value.
        """
        return models.Max(self._name)

    def sum(self):
        """
        Sum of all values in column.
        """
        return models.Sum(self._name)

    #
    # String manipulation
    #
    def lower(self):
        """
        Converts value value to lowercase.
        """
        return functions.Lower(self._name)

    def upper(self):
        """
        Converts value value to uppercase.
        """
        return functions.Upper(self._name)

    def length(self):
        """
        Return the size of string value.
        """
        return functions.Length(self._name)

    def equals(self, value, case=True):
        """
        Compares value with the given value.

        Args:
            value (str of value):
                String value or value expression used for comparison.
            case (bool):
                Set to case=False to perform a case insensitive comparison.
        """
        key = '%s__%s' % (self._name, 'exact' if case else 'iexact')
        return models.Q(**{key: value})

    def regex(self, regex, case=True):
        """
        Match components with a regex.

        Notice that the details of the regex language are backend specific,
        which makes this lookup not portable.

        Args:
            regex (str):
                Regular expression used to match values.
            case (bool):
                Set to case=False to perform a case insensitive match.
        """
        key = '%s__%s' % (self._name, 'regex' if case else 'iregex')
        return models.Q(**{key: regex})

    def startswith(self, prefix, case=True):
        """
        Checks if string string value starts with the given prefix.

        If case=False, performs a case insensitive match.
        """
        key = '%s__%s' % (self._name, 'startswith' if case else 'istartswith')
        return models.Q(**{key: prefix})

    def endswith(self, suffix, case=True):
        """
        Checks if string string value ends with the given suffix.

        If case=False, performs a case insensitive match.
        """
        key = '%s__%s' % (self._name, 'endswith' if case else 'iendswith')
        return models.Q(**{key: suffix})

    def has_substring(self, sub, case=True):
        """
        Checks if string string value has the given substring.

        If case=False, performs a case insensitive search.
        """
        key = '%s__%s' % (self._name, 'contains' if case else 'icontains')
        return models.Q(**{key: sub})

    #
    # Datetimes
    #
    @staticmethod
    def now():
        """
        The current date time.

        Usage:
            Events.objects.filter(F.start_time > F.now())
        """
        return functions.Now()

    year = lookup_property('year')
    month = lookup_property('month')
    day = lookup_property('day')
    week = lookup_property('week')
    week_day = lookup_property('week_day')
    hour = lookup_property('hour')
    minute = lookup_property('minute')
    second = lookup_property('second')

    def extract_from_datetime(self, lookup):
        """
        Extract part of a date, time or datetime by name.

        Usually it is more convenient and safe to use the corresponding
        extractor attribute.

        Usage:
            F.value.extract_from_datetime('year') <==> F.value.year
        """
        return functions.Extract(self._name, lookup_name=lookup)

    def trunc_datetime(self, lookup):
        """
        Truncate datetime in the given component.

        Usually it is more convenient and safe to use the corresponding
        truncation method.

        Usage:
            F.value.trunc_datetime('year') <==> F.value.trunc_year()
        """
        return functions.Trunc(self._name, kind=lookup)

    def trunc_year(self):
        """
        Truncates date or datetime at a whole year.
        """
        return functions.TruncYear(self._name)

    def trunc_month(self):
        """
        Truncates date or datetime at a whole month.
        """
        return functions.TruncMonth(self._name)

    def trunc_day(self):
        """
        Truncates date or datetime at a whole day.
        """
        return functions.TruncDay(self._name)

    def trunc_hour(self):
        """
        Truncates time or datetime at a whole hour.
        """
        return functions.TruncHour(self._name)

    def trunc_minute(self):
        """
        Truncates time or datetime at a whole minute.
        """
        return functions.TruncMinute(self._name)

    def trunc_second(self):
        """
        Truncates time or datetime at a whole second.
        """
        return functions.TruncSecond(self._name)

    def trunc_date(self):
        """
        Truncates datetime to the date component.
        """
        return functions.TruncDate(self._name)

    def trunc_time(self):
        """
        Truncates datetime to the time component.
        """
        return functions.TruncTime(self._name)


#
# Utility functions
#
def setting_attrs(obj, **attrs):
    """
    Save all given keyword arguments as attributes to obj and then return obj.
    """
    for k, v in attrs.items():
        setattr(obj, k, v)
    return obj



import logging

from django.http import HttpResponse

from .utils import not_implemented, allowed_methods, method_map, middleware_chain

log = logging.getLogger('django.request')


class View:
    """
    Boogie class-based views.
    """

    # Default attributes
    template_name = None
    allowed_methods = None
    middleware = ()

    def __init__(self, *, middlewares=(), **kwargs):
        # Init keyword arguments
        cls = type(self)
        for k, v in kwargs.items():
            if hasattr(cls, k):
                setattr(self, k, v)
            else:
                raise TypeError(f'{k} is not an attribute of {cls.__name}')

        self._method_map = method_map(self)
        if self.allowed_methods is None:
            self.allowed_methods = allowed_methods(self)

        self.middlewares = self.select_middlewares(middlewares)
        self._middleware_chain = middleware_chain(self.middlewares, self.request_handler)

    def __call__(self, request, **kwargs):
        return self.respond(request, **kwargs)

    def wrap_method(self, method):
        """
        Prepare a method to respect the view function contract.
        """
        return method

    def select_middlewares(self, middlewares):
        """
        May inject additional middlewares depending on the features requested
        by the view function.
        """
        return middlewares

    def respond(self, request, **kwargs):
        """
        Respond to request.

            view.respond(request, **kwargs) <==> view(request, **kwargs)
        """
        request.view = self
        request.view_args = kwargs
        return self._middleware_chain(request)

    def request_handler(self, request):
        """
        Last point of the middleware chain: receive a request and return a
        response.
        """
        handler = self._method_map[request.method]
        return handler(request, **request.view_args)

    #
    # Http methods
    #
    get = post = delete = put = not_implemented

    def options(self, request, *args, **kwargs):
        """
        Handle responding to requests for the OPTIONS HTTP verb.
        """
        response = HttpResponse()
        methods = map(str.upper, self.allowed_methods)
        response['Allow'] = ', '.join(methods)
        response['Content-Length'] = '0'
        return response



import enum
from functools import lru_cache

from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.translation import ugettext_lazy as _
from sidekick import lazy

NoneType = type(None)


class EnumField(models.Field):
    """
    A Field type that wraps an enumeration.

    It represents each field internally as an integer.

    Args:
        enum_type (type):
            A subclass of :cls:`enum.Enum`. You should consider using
            :cls:`boogie.types.DescriptionEnum` in order to provide
            human-friendly names for each enumeration value.
        is_string (bool):
            If True, data is stored as text in a CharField. The default behavior
            is to store data as integers.
    """

    description = _('An enumeration (using %(is_string)s)')

    @lazy
    def max_abs_value(self):
        return 2 ** 30

    @lazy
    def max_enum_length(self):
        return 10

    def __init__(self, enum_type, *args, is_string=False, **kwargs):
        if not isinstance(enum_type, type) or \
                not issubclass(enum_type, enum.Enum):
            raise ImproperlyConfigured(
                'First argument must be a enum.Enum subclass.'
            )
        if not issubclass(enum_type, enum.IntEnum) and not is_string:
            raise ImproperlyConfigured(
                'Plain enum types are serialized as strings. Either use an '
                'IntEnum enumeration or set is_string=True.'
            )

        self.is_string = is_string
        self.enum_type = enum_type

        # Control maximum length for char fields
        if kwargs.get('max_length', 2 ** 64) < self.max_enum_length:
            raise ImproperlyConfigured(
                'Maximum length is smaller then the larger enum field '
                'representation.'
            )
        if is_string and 'max_length' not in kwargs:
            kwargs['max_length'] = self.max_enum_length

        # Choices
        if 'choices' in kwargs:
            raise ImproperlyConfigured(
                'Cannot set the choices of an enum field.'
            )
        kwargs['choices'] = get_choices_from_enum(enum_type, is_string)

        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        args = [self.enum_type] + args
        del kwargs['choices']
        kwargs['is_string'] = self.is_string
        return name, path, args, kwargs

    def get_internal_type(self):
        if self.is_string:
            return 'CharField'
        elif self.max_abs_value < 32767:
            return 'SmallIntegerField'
        else:
            return 'IntegerField'

    def from_db_value(self, value, expression, connection, context):
        return self.to_python(value)

    def to_python(self, value):
        if isinstance(value, (self.enum_type, NoneType)):
            return value

        if self.is_string or isinstance(value, str):
            return enum_from_str_value(self.enum_type, value)
        else:
            return self.enum_type(int(value))

    def get_db_prep_value(self, *args, **kwargs):
        value = super().get_db_prep_value(*args, **kwargs)

        if value is None:
            return None
        elif self.is_string:
            return str(getattr(value, 'value', value))
        elif isinstance(value, str):
            if value == '':
                return None
            enum = enum_from_str_value(self.enum_type, value)
            return int(enum.value)
        else:
            return int(value)

    def contribute_to_class(self, cls, name, *args, **kwargs):
        super().contribute_to_class(cls, name, *args, **kwargs)
        prefix = name.upper() + '_'
        for option in self.enum_type:
            setattr(cls, prefix + option.name, option)


def get_choices_from_enum(enum_type, is_string=False):
    """
    Return a list of (name, verbose name) choices from an Enum type.
    """

    def human(x):
        return x.lower().replace('_', ' ')

    def description(x):
        try:
            return x.get_description()
        except AttributeError:
            return human(x.name)

    if is_string:
        return tuple((e.value, human(e.name)) for e in enum_type)
    else:
        return tuple((e.value, description(e)) for e in enum_type)


@lru_cache(2048)
def enum_from_str_value(enum_type, value):
    """
    Create Enum instance from a string value. This will scan the list of
    enumerations if string is not found.
    """
    try:
        return enum_type(value)
    except ValueError:
        for enum_value in enum_type:
            if str(enum_value.value) == value:
                return enum_value
        for enum_value in enum_type:
            if enum_value.name == value:
                return enum_value

        raise ValueError('not a valid value for enumeration: %r' % value)



from django.core.exceptions import ImproperlyConfigured
from django.shortcuts import render
from sidekick import lazy
from types import FunctionType


class TemplateMixin:
    """
    Implements the render() and get_context() methods for template-based views.
    """

    render_function: FunctionType = staticmethod(render)
    context_extra: dict = None
    content_type: str = None
    response_status: int = None
    template_engine: str = None
    template_name: str = None
    template_extension: str = '.html'
    template_names: list = property(
        lambda self: [self.template_name] if self.template_name else []
    )

    @lazy
    def view_name(self):
        class_name = self.__class__.__name__
        if class_name.endswith('View'):
            class_name = class_name[:-4]
        # TODO: snake case!
        return class_name.lower()

    def get_template_names(self, request, **kwargs):
        """
        Return a list of template names to look for when rendering the
        template.
        """
        names = self.template_names
        if names:
            return names

        # If no template is given, try to infer it from the model
        model = getattr(self, 'model', None)
        if model is None:
            app_label = model._meta.app_label
            ext = self.template_extension
            return [f'{app_label}/{self.view_name}{ext}']

        # Give up!
        raise ImproperlyConfigured(
            "TemplateResponseMixin requires either a definition of "
            "'template_name' or an implementation of 'get_template_names()'"
        )

    def get_context(self, request, **kwargs):
        """
        Create context dictionary.

        All kwargs are inserted into the dictionary. It also insert a "view"
        variable pointing to the current view and the "view_args" pointing
        to a dictionary with the arguments passed to the view function.
        """
        result = dict(self.context_extra or ())
        result['view'] = self
        result['view_args'] = kwargs
        result.update(kwargs)
        return result

    def render(self, request, **kwargs):
        """
        Return a response, using the `render_function` for this view. Context
        is created with the get_context(request, **kwargs) method.
        """
        context = getattr(request, 'context', None)
        if context is None:
            context = self.get_context(request, **kwargs)

        return self.render_function(
            request,
            self.get_template_names(request, **kwargs),
            context=context,
            content_type=self.content_type,
            status=self.response_status,
            using=self.template_engine,
        )

    def template_middleware(self, next_middleware):
        def middleware(request):
            kwargs = request.view_args
            request.context = self.get_context(request, **kwargs)
            request.render = lambda: self.render(request, **kwargs)

            # We expect a null response or an error. If response is null, we
            # construct our own
            response = next_middleware(request)
            return request.render() if response is None else response

        return middleware



from django.core.exceptions import ImproperlyConfigured
from django.urls import reverse
from rest_framework import serializers

from .utils import join_url


class RestAPISerializer(serializers.HyperlinkedModelSerializer):
    """
    An extended HyperlinkedModelSerializer:

    * Puts all relations to external resources on the obj.links attributes
      (including a self link)
    * Foreign key relations produce links on the links attribute

    Note:
        The outer_relations mapping maps field names of external relations
        (ForeignKeys, OneToOneFields and ManyToManyFields) into a dictionary
        with the following properties::

            {
                'name': <JSON name of resulting external link>,
                'view_name': <View name for relation>,
                'lookup_field': <Field name used for lookups>,
            }
    """
    links = serializers.SerializerMethodField()
    base_name = None
    detail_actions = ()
    list_actions = ()
    outer_relations = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Define base urls
        request = self.context.get('request')
        if request is None:
            self.url_prefix = ''
        else:
            self.url_prefix = f'{request.scheme}://{request.get_host()}/'
        if self.base_name is None:
            cls = type(self)
            name = '%s.%s' % (cls.__module__, cls.__qualname__)
            raise ImproperlyConfigured(
                'Must provide a "base_name" value for the RestAPISerializer '
                'subclass. %s does not define such class attribute.' % name
            )
        base_path = reverse(self.base_name + '-list')
        self.base_url = join_url(self.url_prefix, base_path)

    def self_url(self, obj):
        """
        Return the absolute path (i.e., without the http://host part)
        of the detail url for the current resource.
        """
        # Obtain absolute path
        lookup_field = (
            getattr(self.Meta, 'extra_kwargs', {})
            .get('url', {})
            .get('lookup_field', 'pk')
        )
        kwargs = {lookup_field: getattr(obj, lookup_field)}
        return join_url(self.url_prefix,
                        reverse(self.base_name + '-detail', kwargs=kwargs))

    def get_links(self, obj):
        """
        Return the links dictionary mapping resource names to their
        corresponding links according to HATEAOS.
        """
        return {**self.inner_links(obj), **self.outer_links(obj)}

    def inner_links(self, obj):
        """
        Return a mapping of all inner hyperlinks in object.
        """
        self_url = self.self_url(obj)
        return dict(
            self=self_url,
            **{action: join_url(self_url, action)
               for action in sorted(self.detail_actions)}
        )

    def outer_links(self, obj):
        """
        Return a mapping with all external hyperlinks following from the given
        object.
        """
        if not self.outer_relations:
            return {}
        links = {}
        for attr, relation in self.outer_relations.items():
            value = getattr(obj, attr)
            lookup_field = relation.get('lookup_field', 'pk')
            lookup_value = getattr(value, lookup_field)
            kwargs = {lookup_field: lookup_value}

            name = relation.get('name', attr)
            view_name = relation['view_name']
            links[name] = join_url(self.url_prefix,
                                   reverse(view_name, kwargs=kwargs))
        return links



from django.db.models import functions


def concat(*args):
    """
    Concatenates the values of all given fields or expressions.
    """
    return functions.Concat(*args)


def coalesce(*args):
    """
    Chooses the first non-null value from left to right.

    Usage:
        coalesce(F.savings_account, F.check_account, 0)
    """
    return functions.Coalesce(*args)


def greatest(*args):
    """
    Chooses the greatest of the given values.

    Usage:
        greatest(F.savings_account, F.check_account, 0)
    """
    return functions.Greatest(*args)


def least(*args):
    """
    Chooses the smallest of the given values.

    Usage:
        least(F.savings_account, F.check_account, 0)
    """
    return functions.Least(*args)



from setuptools import setup, find_packages
import sys; sys.path.append('src')


setup(
    name = 'django-boogie',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    setup_requires ='setuptools >= 30.3',
)



from django.db import models
from django.db.models.sql import Query
from sidekick import delegate_to

from .queryset import QuerySet


class Manager(models.Manager):
    """
    Default Boogie manager.
    """

    # Delegates
    update_dataframe = delegate_to('_queryset')
    load_dataframe = delegate_to('_queryset')
    head = delegate_to('head')
    tail = delegate_to('tail')

    def __getitem__(self, item):
        return self.get_queryset().__getitem__(item)

    def __setitem__(self, item, value):
        return self.get_queryset().__setitem__(item, value)

    def get_queryset(self):
        return QuerySet(self.model, Query(self.model), self._db, self._hints)

    _queryset = property(get_queryset)

    def new(self, *args, **kwargs):
        """
        Create a new model instance, without saving it to the database.
        """
        return self.model(*args, **kwargs)

    def save_dataframe(self, *args, **kwargs):
        return self.get_queryset().save_dataframe(*args, **kwargs)


class QueryManager(Manager):
    """
    A manager object constructed from a queryset instance.
    """

    model = delegate_to('queryset')
    name = delegate_to('queryset')
    _db = delegate_to('queryset')
    _hints = delegate_to('queryset')

    def __init__(self, queryset):
        self._queryset = queryset
        self._used_queryset = False

    def __getattr__(self, item):
        return getattr(self._queryset, item)

    def get_queryset(self):
        if self._used_queryset:
            return self._queryset.all()
        else:
            self._used_queryset = True
            return self._queryset



from ..base import Conf


class CompressorConf(Conf):
    # Django compressor
    COMPRESS_ENABLED = True
    COMPRESS_CSS_FILTERS = [
        'compressor.filters.css_default.CssAbsoluteFilter',
        'compressor.filters.cssmin.rCSSMinFilter',
    ]
    COMPRESS_PRECOMPILERS = [
        ('text/coffeescript', 'coffee --compile --stdio'),
        ('text/less', 'lessc {infile} {outfile}'),
        ('text/x-sass', 'sass {infile} {outfile}'),
        ('text/x-scss', 'sass --scss {infile} {outfile}'),
        ('text/stylus', 'stylus < {infile} > {outfile}'),
    ]



from django.apps import apps as _apps

# Import all fields and models. boogie.models serves as a drop-in replacement
# for django models.
from django.db.models import *
from model_utils.models import TimeStampedModel, TimeFramedModel, StatusModel, \
    SoftDeletableModel, SoftDeletableManager, StatusField, AutoCreatedField, \
    MonitorField, AutoLastModifiedField, QueryManager

# Check dependencies before loading custom modules
from .abstract import User, Profile

# Overload Django objects
from .manager import Manager
from .queryset import QuerySet
from .expressions import F, concat, coalesce, greatest, least

# Load Boogie features
from .decorators import manager_method
from ..fields import *

# Django manager utils standalone functions. This is useful for people not using
# boogie managers, but still wants to perform bulk operations
from manager_utils import bulk_update, bulk_upsert, post_bulk_operation, \
    single, upsert, get_or_none, id_dict, sync

# Django pandas functions
from django_pandas.io import read_frame



from copy import copy

from django.db.models import Model, AutoField
from django.utils.functional import cached_property

from .utils import natural_base_url


class ResourceInfo:
    model_name = cached_property(lambda self: self.model.__name__)

    @cached_property
    def field_names(self):
        return [f.name for f in self.model._meta.fields]

    def __init__(self, model: Model,
                 # Fields
                 fields=None, exclude=(),

                 # Urls
                 base_url=None, base_name=None):
        self.model = model
        self.meta = model._meta
        self.model_fields = {f.name: f for f in self.meta.fields}

        # Field info
        self.fields = list(fields or fields_from_model(model))
        for field in exclude:
            self.fields.remove(field)
        self.used_fields = {f for f in self.fields}

        # Url info
        self.base_url = base_url or natural_base_url(model)
        self.base_name = base_name or natural_base_url(model)

        # Hooks
        self.detail_actions = {}
        self.list_actions = {}
        self.properties = {}

    def copy(self):
        """
        Return a copy of itelf.
        """
        return copy(self)

    def add_field(self, name, db_name=None):
        """
        Register a new field name.
        """
        db_name = db_name or name
        self.fields.append(name)

        if name in self.model_fields:
            field = self.model_fields[name]
            self.used_fields_map.append(field)

    def add_property(self, name, method):
        """
        Register a property with the provided name.

        Args:
            name:
                Property name.
            method:
                A function f(obj) -> value that receives a model instance and
                return the corresponding property value.
        """
        self.add_field(name)
        self.properties[name] = method

    def add_action(self, name, method, detail=True):
        """
        Register an action with the given name.
        """
        raise NotImplementedError


def fields_from_model(model):
    return [f.name for f in model._meta.fields if not isinstance(f, AutoField)]



import logging
import operator

from django.utils.functional import cached_property
from sidekick import delegate_to
from rest_framework import serializers
from rest_framework import viewsets

from .api_info import ApiInfo
from .resource_info import ResourceInfo
from .serializers import RestAPISerializer
from .viewsets import RestAPIViewSet

log = logging.getLogger('boogie.rest_api')


class RouterBuilder:
    """
    Builds the router object for the given API and resource.
    """
    # Delegate to API
    version = delegate_to('api')

    # Delegate to resource
    base_url = delegate_to('resource')
    model = delegate_to('resource')
    model_name = delegate_to('resource')
    fields = delegate_to('resource')
    used_fields = delegate_to('resource')

    # Class constants
    viewset_base_class = RestAPIViewSet
    serializer_base_class = RestAPISerializer

    # Computed properties
    @cached_property
    def base_name(self):
        return self.api.get_base_name(self.model)

    def __init__(self, api: ApiInfo, resource: ResourceInfo):
        self.api = api
        self.resource = resource

    def register_at(self, router):
        """
        Register resource in router.

        Args:
            router: a DRF router object.
        """
        viewset = self.create_viewset()
        base_name = self.base_name
        name = viewset.__name__
        log.debug('created viewset %s at %s' % (name, base_name))
        return router.register(self.base_url, viewset, base_name)

    def get_model_field(self, field):
        for f in self.model._meta.fields:
            if f.name == field:
                return f

    #
    # Viewset
    #
    def create_viewset(self):
        bases = (*self.get_viewset_mixins(), self.viewset_base_class)
        ns = self.get_viewset_namespace()
        class_name = self.model_name + 'ViewSet'
        return type(class_name, bases, ns)

    def get_viewset_base_class(self):
        return viewsets.ModelViewSet

    def get_viewset_mixins(self):
        return ()

    def get_viewset_namespace(self):
        model = self.model
        relation_fields = [
            f.name
            for f in model._meta.fields
            if f.related_model is not None and f.name in self.used_fields
        ]

        return dict(
            Meta=self.get_viewset_meta(),
            serializer_class=self.get_serializer_class(),
            queryset=model._default_manager.select_related(*relation_fields),
        )

    def get_viewset_meta(self):
        extra_kwargs = {f: self.get_serializer_extra_kwargs(f) for f in self.fields}

        return dict(
            model=self.model,
            fields=self.fields,
            extra_kwargs=extra_kwargs,
        )

    #
    # Serializer
    #
    def get_serializer_class(self):
        bases = (*self.get_serializer_mixins(), self.serializer_base_class)
        ns = self.get_serializer_namespace()
        class_name = self.model_name + 'Serializer'
        return type(class_name, bases, ns)

    def get_serializer_mixins(self):
        return ()

    def get_serializer_namespace(self):
        api = self.api
        model = self.model
        meta = model._meta
        all_fields = {f.name: f for f in meta.fields}
        extra = {}

        # Outer relations corresponds to the set of all fields that define
        # relations to external models
        outer_relations = {}
        for field_name in self.used_fields:
            field = all_fields[field_name]
            if field.related_model:
                outer_model = field.related_model
                outer_relations[field_name] = dict(
                    name=self.resource.model_fields[field_name].name,
                    view_name=api.get_base_name(outer_model) + '-detail',
                    lookup_url=api.get_lookup_field(outer_model),
                )

        # Method fields are either register externally with the
        # rest_api.attr(Model) decorator or are created implicilty in external
        # relations
        method_field = serializers.SerializerMethodField
        methods = self.get_serializer_method_fields(outer_relations)
        extra.update({'get_' + name: func for name, func in methods.items()})
        extra.update({name: method_field() for name in methods})

        return dict(
            base_name=self.base_name,
            Meta=self.get_serializer_meta(),
            outer_relations=outer_relations,
            **extra,
        )

    def get_serializer_meta(self):
        # Fields is usually just the plain list of fields. The "links" fields
        # is required
        fields = self.fields
        if 'links' not in fields:
            fields = ['links', *fields]

        # Class extra_kwargs attribute
        extra_kwargs = (
            (f, self.get_serializer_extra_kwargs(f))
            for f in self.fields)
        extra_kwargs = {k: v for k, v in extra_kwargs if v}

        # Create a Meta class object for the serializer
        ns = dict(
            model=self.model,
            fields=fields,
            extra_kwargs=extra_kwargs,
        )
        return type('Meta', (), ns)

    def get_serializer_extra_kwargs(self, field):
        extra_kwargs = {}
        field = self.get_model_field(field)

        if field is None:
            return extra_kwargs

        if field.related_model:
            view_name = self.api.get_base_name(field.related_model) + '-detail'
            extra_kwargs['view_name'] = view_name

        return extra_kwargs

    def get_serializer_method_fields(self, outer_relations=None):
        """
        Return a mapping from all method fields to their corresponding
        get_<> methods.
        """
        # Registered method fields
        method_fields = {
            field: staticmethod(method)
            for field, method in self.resource.properties.items()
        }

        # Fields for outer relations
        for field_name, relation in (outer_relations or {}).items():
            name = relation.get('name', field_name)
            getter = operator.attrgetter(field_name)
            method_fields[name] = staticmethod(lambda obj: str(getter(obj)))

        return method_fields



# from django.urls import path


class Router:
    """
    A collection of routes.

    It exports a list of urlpatterns that can be included in a Django's url.py.
    """

    @property
    def patterns(self):
        return []

    def __init__(self):
        self._routes = []
        self._last_url = None

    def __call__(self, url, *args, **kwargs):
        if url is ...:
            url = self._last_url
        self._last_url = url
        append_route = self._routes.append

        if args:
            view_function, = args
            route = Route(url, view_function, **kwargs)
            append_route(route)
            return route
        else:
            # Decorator form
            return lambda view: append_route(Route(url, view)) or view

    #
    # HTTP methods
    #
    def connect(self, *args, **kwargs):
        return self(*args, method='CONNECT', **kwargs)

    def delete(self, *args, **kwargs):
        return self(*args, method='DELETE', **kwargs)

    def get(self, *args, **kwargs):
        return self(*args, method='GET', **kwargs)

    def head(self, *args, **kwargs):
        return self(*args, method='HEAD', **kwargs)

    def options(self, *args, **kwargs):
        return self(*args, method='OPTIONS', **kwargs)

    def patch(self, *args, **kwargs):
        return self(*args, method='PATCH', **kwargs)

    def post(self, *args, **kwargs):
        return self(*args, method='POST', **kwargs)

    def put(self, *args, **kwargs):
        return self(*args, method='PUT', **kwargs)

    def trace(self, *args, **kwargs):
        return self(*args, method='TRACE', **kwargs)


class Route:
    """
    Represents a Boogie route.
    """

    def __init__(self, path, function, name=None, template=None, login=False, perms=None, method=None):
        self.path = as_path(path)
        self.function = function
        self.name = name
        self.template = template
        self.login = login
        self.perms = perms
        self.method = method


class Path:
    """
    A path that uses an extended Django path.
    """

    def __init__(self, path):
        self.string = path


def as_path(path):
    return path if isinstance(path, Path) else Path(path)



class Db:
    """
    Provides easy access to model tables.
    """

    def __init__(self, app=None):
        self._app = app

    def __getattr__(self, attr):
        if self._app is None:
            return Db(attr)
        else:
            from django.apps import apps

            model = apps.get_model(self._app, attr)
            return wrap_as_boogie_manager(model._default_manager)


db = Db()


def wrap_as_boogie_manager(manager):
    """
    Wraps a manager instance as a Boogie-compatible manager.
    """
    # TODO
    return manager



from django.core.exceptions import ImproperlyConfigured

try:
    import rest_framework as _rest
except ImportError:
    raise ImproperlyConfigured(
        'boogie.rest requires Django Rest Framework and some additional '
        'packages to be installed. Please install it using pip install '
        'django-boogie[rest].'
    )

from .base import RestAPI

rest_api = RestAPI()
